"""
Test that the current version of Starsim exactly matches
the baseline results.
"""

import numpy as np
import sciris as sc
import starsim as ss

baseline_filename  = sc.thisdir(__file__, 'baseline.json')
multisim_filename  = sc.thisdir(__file__, 'baseline_multisim.json')
benchmark_filename = sc.thisdir(__file__, 'benchmark.json')
parameters_filename = sc.thisdir(ss.__file__, 'regression', f'pars_v{ss.__version__}.json')
sc.options(interactive=False) # Assume not running interactively

# Define the parameters
pars = sc.objdict(
    n_agents  = 10e3, # Number of agents
    start     = 2000, # Starting year
    dur       = 20,   # Number of years to simulate
    dt        = 0.2,  # Timestep
    verbose   = 0,    # Don't print details of the run
    rand_seed = 2,    # Set a non-default seed
)


def make_sim(run=False, **kwargs):
    """
    Define a default simulation for testing the baseline. If run directly (not
    via pytest), also plot the sim by default.
    """
    diseases = ['sir', 'sis']
    networks = ['random', 'mf', 'maternal']
    sim = ss.Sim(pars=pars | kwargs, networks=networks, diseases=diseases, demographics=True)

    # Optionally run and plot
    if run:
        sim.run()
        sim.plot()

    return sim


def save_baseline():
    """
    Refresh the baseline results. This function is not called during standard testing,
    but instead is called by the update_baseline script.
    """
    sc.heading('Updating baseline values...')

    # Make and run sim
    sim = make_sim()
    sim.run()

    # Export results
    sim.to_json(filename=baseline_filename, keys='summary')

    # Save parameters
    sim.to_json(filename=parameters_filename, keys='pars') # If not different from previous version, can safely delete

    print('Done.')
    return


def multisim_baseline(save=False, n_runs=10, **kwargs):
    """
    Check or update the multisim baseline results; not part of the integration tests
    """
    word = 'Saving' if save else 'Checking'
    sc.heading(f'{word} MultiSim baseline...')

    # Make and run sims
    kwargs.setdefault('verbose', ss.options.verbose/n_runs*2)
    sim = make_sim(**kwargs)
    msim = ss.MultiSim(base_sim=sim)
    msim.run(n_runs)
    summary = msim.summarize()

    # Export results
    if save:
        sc.savejson(multisim_filename, summary)
        print('Done.')
    else:
        baseline = sc.loadjson(multisim_filename)
        ss.diff_sims(baseline, summary, die=False)

    return


def test_baseline():
    """ Compare the current default sim against the saved baseline """

    # Load existing baseline
    baseline = sc.loadjson(baseline_filename)
    old = baseline['summary']

    # Calculate new baseline
    new = make_sim()
    new.run()

    # Compute the comparison
    ss.diff_sims(old, new, die=True)

    return new


def test_benchmark(do_save=False, repeats=1, verbose=True):
    """ Compare benchmark performance """

    if verbose: print('Running benchmark...')
    try:
        previous = sc.loadjson(benchmark_filename)
    except FileNotFoundError:
        previous = None

    t_inits = []
    t_runs  = []

    def normalize_performance():
        """ Normalize performance across CPUs """
        t_bls = []
        bl_repeats = 3
        n_outer = 10
        n_inner = 1e6
        for r in range(bl_repeats):
            t0 = sc.tic()
            for i in range(n_outer):
                a = np.random.random(int(n_inner))
                b = np.random.random(int(n_inner))
                a*b
            t_bl = sc.toc(t0, output=True)
            t_bls.append(t_bl)
        t_bl = min(t_bls)
        reference = 0.07 # Benchmarked on an Intel i7-12700H CPU @ 2.90GHz
        ratio = reference/t_bl
        return ratio


    # Test CPU performance before the run
    r1 = normalize_performance()

    # Do the actual benchmarking
    for r in range(repeats):

        print(f'Repeat {r}')

        # Time initialization
        t0 = sc.tic()
        sim = make_sim()
        sim.init()
        t_init = sc.toc(t0, output=True)

        # Time running
        t0 = sc.tic()
        sim.run()
        t_run = sc.toc(t0, output=True)

        # Store results
        t_inits.append(t_init)
        t_runs.append(t_run)

    # Test CPU performance after the run
    r2 = normalize_performance()
    ratio = (r1+r2)/2
    t_init = ratio*min(t_inits)
    t_run  = ratio*min(t_runs)

    # Construct json
    n_decimals = 3
    json = {'time': {
                'initialize': round(t_init, n_decimals),
                'run':        round(t_run,  n_decimals),
                },
            'parameters': {
                'n_agents': sim.pars.n_agents,
                'dur':      sim.pars.dur,
                'dt':       sim.pars.dt,
                },
            'cpu_performance': ratio,
            }

    if verbose:
        if previous:
            print('Previous benchmark:')
            sc.pp(previous)

        print('\nNew benchmark:')
        sc.pp(json)
    else:
        brief = sc.dcp(json['time'])
        brief['cpu_performance'] = json['cpu_performance']
        sc.pp(brief)

    if do_save:
        sc.savejson(filename=benchmark_filename, obj=json, indent=2)

    if verbose:
        print('Done.')

    return json


if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)
    T = sc.timer()

    json = test_benchmark() # Run this first so benchmarking is available even if results are different
    new  = test_baseline()
    sim = make_sim(run=do_plot)

    T.toc()



"""
Test different time units and timesteps
"""

import numpy as np
import sciris as sc
import starsim as ss

small = 100
medium = 1000
sc.options(interactive=False)


# %% Define the tests
def test_ratio():
    sc.heading('Test behavior of time_ratio()')

    assert ss.time_ratio() == 1.0
    assert ss.time_ratio(dt1=1.0, dt2=0.1) == 10.0
    assert ss.time_ratio(dt1=0.5, dt2=5) == 0.1
    assert ss.time_ratio(dt1=0.5, dt2=5) == 0.1

    assert ss.time_ratio(unit1='year', unit2='day') == 365.25
    assert ss.time_ratio(unit1='day', unit2='year') == 1/365.25

    return


def test_classes():
    sc.heading('Test behavior of dur() and rate()')

    # Test duration dt
    d1 = ss.dur(2)
    d2 = ss.dur(3)
    d3 = ss.dur(2, parent_dt=0.1)
    d4 = ss.dur(3, parent_dt=0.2)
    for d in [d1,d2,d3,d4]: d.init()

    assert d1 + d2 == 2+3
    assert d3 + d4 == 2/0.1 + 3/0.2
    assert d3 * 2 == 2/0.1*2
    assert d3 / 2 == 2/0.1/2

    # Test rate dt
    r1 = ss.rate(2)
    r2 = ss.rate(3)
    r3 = ss.rate(2, parent_dt=0.1)
    r4 = ss.rate(3, parent_dt=0.2)
    for r in [r1,r2,r3,r4]: r.init()

    assert r1 + r2 == 2 + 3
    assert r3 + r4 == 2*0.1 + 3*0.2
    assert r3 * 2 == 2*0.1*2
    assert r3 / 2 == 2*0.1/2

    # Test duration units
    d5 = ss.dur(2, unit='year').init(parent_unit='day')
    d6 = ss.dur(3, unit='day').init(parent_unit='day')
    assert d5 + d6 == 2*365.25+3

    # Test rate units
    rval = 0.7
    r5 = ss.rate(rval, unit='week').init(parent_unit='day')
    r6 = ss.rate(rval).init(parent_dt=0.1)
    assert np.isclose(r5.values, rval/7) # A limitation of this approach, not exact!
    assert np.isclose(r6.values, rval/10)

    # Test time_prob
    tpval = 0.1
    tp0 = ss.time_prob(tpval).init(parent_dt=1.0)
    tp1 = ss.time_prob(tpval).init(parent_dt=0.5)
    tp2 = ss.time_prob(tpval).init(parent_dt=2)
    assert np.isclose(tp0.values, tpval)
    assert np.isclose(tp1.values, tpval/2, rtol=0.1)
    assert np.isclose(tp2.values, tpval*2, rtol=0.1)
    assert tp1.values > tpval/2
    assert tp2.values < tpval*2

    return d3, d4, r3, r4, tp1


def test_units(do_plot=False):
    sc.heading('Test behavior of year vs day units')

    sis = ss.SIS(
        beta = ss.beta(0.05, 'day'),
        init_prev = ss.bernoulli(p=0.1),
        dur_inf = ss.lognorm_ex(mean=ss.dur(10, 'day')),
        waning = ss.rate(0.05, 'day'),
        imm_boost = 1.0,
    )

    rnet = ss.RandomNet(
        n_contacts = 10,
        dur = 0, # Note; network edge durations are required to have the same unit as the network
    )

    pars = dict(
        diseases = sis,
        networks = rnet,
        n_agents = small,
    )

    sims = sc.objdict()
    sims.y = ss.Sim(pars, unit='year', label='Year', start=2000, stop=2002, dt=1/365, verbose=0)
    sims.d = ss.Sim(pars, unit='day', label='Day', start='2000-01-01', stop='2002-01-01', dt=1, verbose=0)

    for sim in sims.values():
        sim.run()
        if do_plot:
            sim.plot()

    # Check that results match to within stochastic uncertainty
    rtol = 0.05
    vals = [sim.summary.sis_cum_infections for sim in [sims.y, sims.d]]
    assert np.isclose(*vals, rtol=rtol), f'Values for cum_infections do not match ({vals})'

    return sims


def test_multi_timestep(do_plot=False):
    sc.heading('Test behavior of different modules having different timesteps')

    pars = dict(
        diseases = ss.SIS(unit='day', dt=1.0, init_prev=0.1, beta=ss.beta(0.01)),
        demographics = ss.Births(unit='year', dt=0.25),
        networks = ss.RandomNet(unit='week'),
        n_agents = small,
        verbose = 0,
    )

    sim = ss.Sim(pars, unit='day', dt=2, start='2000-01-01', stop='2002-01-01')
    sim.run()

    twoyears = 366*2
    quarters = 2/0.25
    assert len(sim) == twoyears//2
    assert len(sim.diseases.sis) == twoyears
    assert len(sim.demographics.births) == quarters

    if do_plot:
        sim.plot()

    return sim


def test_mixed_timesteps():
    sc.heading('Test behavior of different commbinations of timesteps')

    siskw = dict(dur_inf=ss.dur(50, 'day'), beta=ss.beta(0.01, 'day'), waning=ss.rate(0.005, 'day'))
    kw = dict(n_agents=1000, start='2001-01-01', stop='2001-07-01', networks='random', copy_inputs=False, verbose=0)

    print('Year-year')
    sis1 = ss.SIS(unit='year', dt=1/365, **sc.dcp(siskw))
    sim1 = ss.Sim(unit='year', dt=1/365, diseases=sis1, label='year-year', **kw)

    print('Day-day')
    sis2 = ss.SIS(unit='day', dt=1.0, **sc.dcp(siskw))
    sim2 = ss.Sim(unit='day', dt=1.0, diseases=sis2, label='day-day', **kw)

    print('Day-year')
    sis3 = ss.SIS(unit='day', dt=1.0, **sc.dcp(siskw))
    sim3 = ss.Sim(unit='year', dt=1/365, diseases=sis3, label='day-year', **kw)

    print('Year-day')
    sis4 = ss.SIS(unit='year', dt=1/365, **sc.dcp(siskw))
    sim4 = ss.Sim(unit='day', dt=1.0, diseases=sis4, label='year-day', **kw)

    msim = ss.parallel(sim1, sim2, sim3, sim4)

    # Check that al results are close
    threshold = 0.01
    summary = msim.summarize()
    for key,res in summary.items():
        if res.mean:
            ratio = res.std/res.mean
            assert ratio < threshold, f'Result {key} exceeds threshold: {ratio:n} > {threshold}'
            print(f'✓ Result {key} within threshold: {ratio:n} < {threshold}')
    return msim


def test_time_class():
    sc.heading('Test different instances of ss.Time')

    def sim(start, stop, dt, unit):
        """ Generate a fake sim """
        sim = sc.prettyobj()
        sim.t = ss.Time(start=start, stop=stop, dt=dt, unit=unit, sim=True)
        return sim

    print('Testing dates vs. numeric')
    s1 = sim(start=2000, stop=2002, dt=0.1, unit='year')
    t1 = ss.Time(start='2001-01-01', stop='2001-06-30', dt=2.0, unit='day')
    t1.init(sim=s1)
    assert np.array_equal(s1.t.timevec, s1.t.yearvec)
    assert len(s1.t.timevec) == 21
    assert t1.npts == sc.daydiff('2001-01-01', '2001-06-30')//2 + 1
    assert sc.isnumber(s1.t.start)
    assert isinstance(t1.start, ss.date)
    assert s1.t.datevec[-1] == ss.date('2002-01-01')
    assert t1.datevec[-1] == ss.date('2001-06-30')
    assert t1.abstvec[0] == 1.0

    print('Testing weeks vs. days')
    s2 = sim(start='2000-06-01', stop='2001-05-01', dt=1, unit='day')
    t2 = ss.Time(start='2000-06-01', stop='2001-05-01', dt=1, unit='week')
    t2.init(sim=s2)
    assert np.array_equal(s2.t.timevec, s2.t.datevec)
    assert isinstance(s2.t.start, ss.date)
    assert t2.npts*7 == s2.t.npts + 1
    assert np.array_equal(t2.abstvec, s2.t.abstvec[::7])

    print('Testing different units and dt')
    s3 = sim(start='2001-01-01', stop='2003-01-01', dt=0.1, unit='year')
    t3 = ss.Time(start='2001-01-01', stop='2003-01-01', dt=2.0, unit='day')
    t3.init(sim=s3)
    assert np.array_equal(s3.t.timevec, s3.t.datevec)
    assert s3.t.datevec[-1] == ss.date('2003-01-01')
    assert s3.t.npts == 21
    assert np.isclose(s3.t.abstvec.mean(), t3.abstvec.mean(), atol=1e-3)

    print('Testing unitless')
    s4 = sim(start=0, stop=10, dt=1.0, unit='unitless')
    t4 = ss.Time(start=2, stop=9, dt=0.1, unit='unitless')
    t4.init(sim=s4)
    assert np.array_equal(s4.t.timevec, s4.t.abstvec)
    assert s4.t.datevec[0] == ss.date(ss.time.default_start_year)
    assert len(s4.t) == 11
    assert len(t4) == (9-2)/0.1+1

    return [s1, t1, s2, t2]


# %% Run as a script
if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)

    T = sc.timer()

    o1 = test_ratio()
    o2 = test_classes()
    o3 = test_units(do_plot)
    o4 = test_multi_timestep(do_plot)
    o5 = test_mixed_timesteps()
    o6 = test_time_class()

    T.toc()



"""
Test distributions
"""

# %% Imports and settings
import starsim as ss
import matplotlib.pyplot as plt
import numpy as np

data = np.random.normal(size=1000)
hist, bins = np.histogram(data, bins=50)
plt.subplot(121)
plt.hist(data, 50)
plt.subplot(122)
plt.hist(ss.data_dist(vals=hist, bins=bins)._sample(10_000), 50)
plt.show()



import sciris as sc
import starsim as ss

kw = dict(n_agents=10e3, start=2000, dur=100, diseases='sis', networks='random', demographics=True)
sim = ss.Sim(**kw)

prof = ['profile', 'cprofile', 'time'][1]

if prof == 'profile':
    prf = sc.profile(run=sim.run, follow=[sim.run])

elif prof == 'cprofile':
    cprf = sc.cprofile(sort='selfpct', mintime=1e-3, maxitems=None, maxfunclen=None, maxpathlen=None)
    with cprf:
        sim.run()

else:
    with sc.timer() as T:
        sim.run()

sim.loop.plot_cpu()


import numpy as np
import sciris as sc
import pylab as pl

n = 100

reps = 1001
pause = 0.5
nframes = 20
nbins = n

all_diffs = sc.objdict(Random=[], Modulo=[])
sum_diffs = sc.objdict(Random=[], Modulo=[])

# Choose logarithmically spaced frames to plot
frames = np.linspace(0, np.log(reps), nframes)
frames = np.exp(frames).astype(int) - 1

sc.options(dpi=200)
fig = pl.figure(figsize=(12,8))
sc.figlayout()

for rep in range(reps):
    a = np.random.rand(n,n)
    b1 = np.random.rand(n,1)
    b2 = np.random.rand(1,n)
    B1 = np.tile(b1, (1,n))
    B2 = np.tile(b2, (n,1))
    b = (B1 + B2) % 1.0
    
    arrdict = sc.objdict(Random=a, Modulo=b)
    
    for i,key,arr in arrdict.enumitems():
    
        v1 = np.diff(arr, axis=0)
        v2 = np.diff(arr, axis=1)
        v = sc.cat(v1.flatten(), v2.flatten()).tolist()
        all_diffs[key] += v
        
        sv = np.diff(arr.sum(axis=0)).tolist()
        sum_diffs[key] += sv
    
    if rep in frames:
        fig.clear()
        for i,key,arr in arrdict.enumitems():
            
            pl.subplot(2,3,1+i*3)
            pl.imshow(arr, cmap='turbo')
            t = f'{key}: rep={rep}\nmean={arr.mean():0.3f}, std={arr.std():0.3f}\nmin={arr.min():0.3f}, max={arr.max():0.3f}'
            pl.title(t)
            
            pl.subplot(2,3,2+i*3)
            hist, bins = np.histogram(all_diffs[key], bins=nbins)
            w = np.diff(bins).mean()
            pl.bar(bins[:-1], hist, width=w, facecolor='k', lw=0)
            pl.title(f'{key}: all diffs rep={rep}')
            sc.commaticks()
            
            pl.subplot(2,3,3+i*3)
            shist, sbins = np.histogram(sum_diffs[key], bins=nbins)
            w = np.diff(sbins).mean()
            pl.bar(sbins[:-1], shist, width=w, facecolor='k', lw=0)
            pl.title(f'{key}: sum diffs rep={rep}')
            sc.commaticks()
        
        pl.pause(pause)



"""
Run tests of vaccines
"""

# %% Imports and settings
import sciris as sc
import numpy as np
import starsim as ss


def run_sir_vaccine(efficacy, leaky=True, do_plot=False):
    # parameters
    v_frac      = 0.5    # fraction of population vaccinated
    total_cases = 500    # total cases at which point we check results
    tol         = 3      # tolerance in standard deviations for simulated checks

    # create a basic SIR sim
    sim = ss.Sim(
        n_agents = 1000,
        pars = dict(
          networks = dict(
                type = 'random',
                n_contacts = 4
          ),
          diseases = dict(
                type      = 'sir',
                init_prev = 0.01,
                dur_inf   = ss.dur(10),
                p_death   = 0,
                beta      = ss.beta(0.06),
          )
        ),
        dur = 10,
        dt  = 0.01
    )
    sim.init(verbose=False)

    # work out who to vaccinate
    in_trial = sim.people.sir.susceptible.uids
    n_vac = round(len(in_trial) * v_frac)
    in_vac = np.random.choice(in_trial, n_vac, replace=False)
    in_pla = np.setdiff1d(in_trial, in_vac)
    uids = ss.uids(in_vac)

    # create and apply the vaccination
    vac = ss.sir_vaccine(efficacy=efficacy, leaky=leaky)
    vac.init_pre(sim)
    vac.administer(sim.people, uids)

    # check the relative susceptibility at the start of the simulation
    rel_susc = sim.people.sir.rel_sus.values
    assert min(rel_susc[in_pla]) == 1.0, 'Placebo arm is not fully susceptible'
    if not leaky:
        assert min(rel_susc[in_vac]) == 0.0, 'Nobody fully vaccinated (all_or_nothing)'
        assert max(rel_susc[in_vac]) == 1.0, 'Vaccine effective in everyone (all_or_nothing)'
        mean = n_vac * (1 - efficacy)
        sd = np.sqrt(n_vac * efficacy * (1 - efficacy))
        assert (np.mean(rel_susc[in_vac]) - mean) / sd < tol, 'Incorrect mean susceptibility in vaccinated (all_or_nothing)'
    else:
        assert max(abs(rel_susc[in_vac] - (1 - efficacy))) < 0.0001, 'Relative susceptibility not 1-efficacy (leaky)'

    # run the simulation until sufficient cases
    old_cases = []
    for idx in range(1000):
        sim.run_one_step()
        susc = sim.people.sir.susceptible.uids
        cases = np.setdiff1d(in_trial, susc)
        if len(cases) > total_cases:
            break
        old_cases = cases

    if len(cases) > total_cases:
        cases = np.concatenate([old_cases, np.random.choice(np.setdiff1d(cases, old_cases), total_cases - len(old_cases), replace=False)])
    vac_cases = np.intersect1d(cases, in_vac)

    # check to see whether the number of cases are as expected
    p = v_frac * (1 - efficacy) / (1 - efficacy * v_frac)
    mean = total_cases * p
    sd = np.sqrt(total_cases * p * (1 - p))
    assert (len(vac_cases) - mean) / sd < tol, 'Incorrect proportion of vaccincated infected'

    # for all or nothing check that fully vaccinated did not get infected
    if not leaky:
        assert len(np.intersect1d(vac_cases, in_vac[rel_susc[in_vac] == 1.0])) == len(vac_cases), 'Not all vaccine cases amongst vaccine failures (all or nothing)'
        assert len(np.intersect1d(vac_cases, in_vac[rel_susc[in_vac] == 0.0])) == 0, 'Vaccine cases amongst fully vaccincated (all or nothing)'

    if do_plot:
        sim.plot()

    return sim


def test_sir_vaccine_leaky(do_plot=False):
    return run_sir_vaccine(0.3, False, do_plot=do_plot)

def test_sir_vaccine_all_or_nothing(do_plot=False):
    return run_sir_vaccine(0.3, True, do_plot=do_plot)


if __name__ == '__main__':
    T = sc.timer()
    do_plot = True

    leaky  = test_sir_vaccine_leaky(do_plot=do_plot)
    a_or_n = test_sir_vaccine_all_or_nothing(do_plot=do_plot)

    T.toc()



"""
Define non-communicable disease (NCD) model
"""

import numpy as np
import starsim as ss
import sciris as sc


__all__ = ['NCD']

class NCD(ss.Disease):
    """
    Example non-communicable disease

    This class implements a basic NCD model with risk of developing a condition
    (e.g., hypertension, diabetes), a state for having the condition, and associated
    mortality.
    """
    def __init__(self, pars=None, **kwargs):
        super().__init__()
        self.define_pars(
            initial_risk = ss.bernoulli(p=0.3), # Initial prevalence of risk factors
            dur_risk = ss.expon(scale=ss.dur(10)),
            prognosis = ss.weibull(c=ss.years(2), scale=5), # Time in years between first becoming affected and death
        )
        self.update_pars(pars=pars, **kwargs)

        self.define_states(
            ss.State('at_risk', label='At risk'),
            ss.State('affected', label='Affected'),
            ss.FloatArr('ti_affected', label='Time of becoming affected'),
            ss.FloatArr('ti_dead', label='Time of death'),
        )
        return

    @property
    def not_at_risk(self):
        return ~self.at_risk

    def init_post(self):
        """
        Set initial values for states. This could involve passing in a full set of initial conditions,
        or using init_prev, or other. Note that this is different to initialization of the State objects
        i.e., creating their dynamic array, linking them to a People instance. That should have already
        taken place by the time this method is called.
        """
        super().init_post()
        initial_risk = self.pars['initial_risk'].filter()
        self.at_risk[initial_risk] = True
        self.ti_affected[initial_risk] = self.ti + sc.randround(self.pars['dur_risk'].rvs(initial_risk))
        return initial_risk

    def step_state(self):
        ti = self.ti
        deaths = (self.ti_dead == ti).uids
        self.sim.people.request_death(deaths)
        if self.pars.log:
            self.log.add_data(deaths, died=True)
        self.results.new_deaths[ti] = len(deaths) # Log deaths attributable to this module
        return

    def step(self):
        ti = self.ti
        new_cases = (self.ti_affected == ti).uids
        self.affected[new_cases] = True
        prog_years = self.pars.prognosis.rvs(new_cases)
        self.ti_dead[new_cases] = ti + sc.randround(prog_years / self.t.dt) # TODO: update to allow non-year units
        super().set_prognoses(new_cases)
        return new_cases

    def init_results(self):
        """
        Initialize results
        """
        super().init_results()
        self.define_results(
            ss.Result('n_not_at_risk', dtype=int,   label='Not at risk'),
            ss.Result('prevalence',    dtype=float, label='Prevalence'),
            ss.Result('new_deaths',    dtype=int,   label='Deaths'),
        )
        return

    def update_results(self):
        super().update_results()
        ti = self.ti
        self.results.n_not_at_risk[ti] = np.count_nonzero(self.not_at_risk)
        self.results.prevalence[ti]    = np.count_nonzero(self.affected)/len(self.sim.people)
        self.results.new_deaths[ti]    = np.count_nonzero(self.ti_dead == ti)
        return



"""
Result structures.
"""
import numpy as np
import sciris as sc
import starsim as ss
import matplotlib.pyplot as plt

__all__ = ['Result', 'Results']


class Result(ss.BaseArr):
    """
    Array-like container for holding sim results.

    Args:
        module (str): the name of the parent module, e.g. 'hiv'
        name (str): the name of this result, e.g. 'new_infections'
        shape (int/tuple): the shape of the result array (usually module.npts)
        scale (bool): whether or not the result scales by population size (e.g. a count does, a prevalence does not)
        label (str): a human-readable label for the result
        values (array): prepopulate the Result with these values
        timevec (array): an array of time points
        low (array): values for the lower bound
        high (array): values for the upper bound

    In most cases, ``ss.Result`` behaves exactly like ``np.array()``, except with
    the additional fields listed above. To see everything contained in a result,
    you can use result.disp().
    """
    def __init__(self, name=None, label=None, dtype=float, shape=None, scale=True,
                 module=None, values=None, timevec=None, low=None, high=None):
        # Copy inputs
        self.name = name
        self.label = label
        self.module = module
        self.scale = scale
        self.timevec = timevec
        self.low = low
        self.high = high
        self.dtype = dtype
        self.shape = shape
        self.values = values
        self.init_values()
        return

    def __repr__(self):
        cls_name = self.__class__.__name__
        arrstr = super().__repr__().removeprefix(cls_name)
        out = f'{cls_name}({self.key}):\narray{arrstr}'
        return out

    def __str__(self):
        cls_name = self.__class__.__name__
        try:
            minval = self.values.min()
            meanval = self.values.mean()
            maxval = self.values.max()
            valstr = f'min={minval:n}, mean={meanval:n}, max={maxval:n}'
        except:
            valstr = f'{self.values}'
        out = f'{cls_name}({self.key}: {valstr})'
        return out

    def __getitem__(self, key):
        """ Allow e.g. result['low'] """
        if isinstance(key, str):
            return getattr(self, key)
        else:
            return super().__getitem__(key)

    @property
    def initialized(self):
        return self.values is not None

    @property
    def has_dates(self):
        """ Check whether the time vector uses dates (rather than numbers) """
        try:    return not sc.isnumber(self.timevec[0])
        except: return False

    def init_values(self, values=None, dtype=None, shape=None, force=False):
        """ Handle values """
        if not self.initialized or force:
            values = sc.ifelse(values, self.values)
            dtype = sc.ifelse(dtype, self.dtype)
            shape = sc.ifelse(shape, self.shape)
            if values is not None: # Create if values already supplied
                self.values = np.array(values, dtype=dtype)
                dtype = self.values.dtype
                shape = self.values.shape
            elif shape is not None: # Or if a shape is provided, initialize
                self.values = np.zeros(shape=shape, dtype=dtype)
            else:
                self.values = None
            self.dtype = dtype
            self.shape = shape
        return self.values

    def update(self, *args, **kwargs):
        """ Update parameters, and initialize values if needed """
        super().update(*args, **kwargs)
        self.init_values()
        return

    @property
    def key(self):
        """ Return the unique key of the result: <module>.<name>, e.g. "hiv.new_infections" """
        modulestr = f'{self.module}.' if (self.module is not None) else ''
        namestr = self.name if (self.name is not None) else 'unnamed'
        key = modulestr + namestr
        return key

    @property
    def full_label(self):
        """ Return the full label of the result: <Module>: <label>, e.g. "HIV: New infections" """
        reslabel = sc.ifelse(self.label, self.name)
        if self.module == 'sim': # Don't add anything if it's the sim
            full = f'Sim: {reslabel}'
        else:
            try:
                mod = ss.find_modules(flat=True)[self.module]
                modlabel = mod.__name__
                assert self.module == modlabel.lower(), f'Mismatch: {self.module}, {modlabel}' # Only use the class name if the module name is the default
            except: # Don't worry if we can't find it, just use the module name
                modlabel = self.module.title()
            full = f'{modlabel}: {reslabel}'
        return full

    def to_df(self, sep='_', rename=False):
        """
        Convert to a dataframe with timevec, value, low, and high columns

        Args:
            rename (bool): if True, rename the columns with the name of the result (else value, low, high)
        """
        data = dict()
        if self.timevec is not None:
            data['timevec'] = self.timevec
        valcol = self.name if rename else 'value'
        data[valcol] = self.values
        for key in ['low', 'high']:
            val = self[key]
            valcol = f'{self.name}{sep}{key}' if rename else key
            if val is not None:
                data[valcol] = val
        df = sc.dataframe(data)
        return df

    def plot(self, fig=None, ax=None, fig_kw=None, plot_kw=None, fill_kw=None, **kwargs):
        """ Plot a single result; kwargs are interpreted as plot_kw """
        # Prepare inputs
        fig_kw = sc.mergedicts(fig_kw)
        plot_kw = sc.mergedicts(dict(lw=3, alpha=0.8), plot_kw, kwargs)
        fill_kw = sc.mergedicts(dict(alpha=0.1), fill_kw)
        if fig is None and ax is None:
            fig = plt.figure(**fig_kw)
        if ax is None:
            ax = plt.subplot(111)
        if self.timevec is None:
            errormsg = f'Cannot figure out how to plot {self}: no time data associated with it'
            raise ValueError(errormsg)

        # Plot bounds
        if self.low is not None and self.high is not None:
            ax.fill_between(self.timevec, self.low, self.high, **fill_kw)

        # Plot results
        plt.plot(self.timevec, self.values, **plot_kw)
        plt.title(self.full_label)
        plt.xlabel('Time')
        sc.commaticks(ax)
        if self.has_dates:
            sc.dateformatter(ax)
        if (self.values.min() >= 0) and (plt.ylim()[0]<0): # Don't allow axis to go negative if results don't
            plt.ylim(bottom=0)
        return fig


class Results(ss.ndict):
    """ Container for storing results """
    def __init__(self, module, *args, strict=True, **kwargs):
        if hasattr(module, 'name'):
            module = module.name
        self.setattribute('_module', module)
        super().__init__(type=Result, strict=strict, *args, **kwargs)
        return

    def __repr__(self, indent=2, **kwargs): # kwargs are not used, but are needed for disp() to work
        string = f'Results({self._module})\n'
        for i,k,v in self.enumitems():
            if k != 'timevec':
                entry = f'{v}'
                if '\n' in entry: # Check if the string is multi-line
                    lines = entry.splitlines()
                    entry = f'{i}. {lines[0]}\n'
                    entry += '\n'.join(' '*indent + f'{i}.' + line for line in lines[1:])
                    string += entry + '\n'
                else:
                    string += f'{i}. {v}\n'
        string = string.rstrip()
        return string

    def disp(self, *args, **kwargs):
        print(super().__repr__(*args, **kwargs))
        return

    def append(self, arg, key=None):
        """ This is activated by adding as well, e.g. results += result """
        if isinstance(arg, (list, tuple)):
            result = ss.Result(self._module, *arg)
        elif isinstance(arg, dict):
            result = ss.Result(self._module, **arg)
        else:
            result = arg

        if not isinstance(result, Result):
            warnmsg = f'You are adding a result of type {type(result)} to Results, which is inadvisable.'
            ss.warn(warnmsg)

        if result.module != self._module:
            if result.module:
                warnmsg = f'You are adding a result from module {result.module} to module {self._module}; check that this is intentional.'
                ss.warn(warnmsg)
            result.module = self._module

        super().append(result, key=key)
        return

    @property
    def all_results(self):
        """ Iterator over all results, skipping any nested values """
        return iter(res for res in self.values() if isinstance(res, Result))

    @property
    def equal_len(self):
        """ Check if all results are equal length """
        lengths = [len(res) for res in self.flatten().values()]
        return len(set(lengths)) == 1

    def flatten(self, sep='_', only_results=True):
        """ Turn from a nested dictionary into a flat dictionary, keeping only results by default """
        out = sc.flattendict(self, sep=sep)
        if only_results:
            out = sc.objdict({k:v for k,v in out.items() if isinstance(v, Result)})
        return out

    def to_df(self, sep='_', descend=False):
        """ Merge all results dataframes into one """
        if not descend:
            dfs = [res.to_df(sep=sep, rename=True) for res in self.all_results]
            if len(dfs):
                df = dfs[0]
                for df2 in dfs[1:]:
                    df = df.merge(df2)
            else:
                df = None
        else:
            if self.equal_len:
                flat = self.flatten(sep=sep, only_results=True)
                flat = dict(timevec=self.timevec) | flat # Prepend the timevec
                df = sc.dataframe.from_dict(flat)
            else:
                df = sc.objdict() # For non-equal lengths, actually return an objdict rather than a dataframe
                df.sim = self.to_df(sep=sep, descend=False)
                for k,v in self.items():
                    if isinstance(v, Results):
                        thisdf = v.to_df(sep=sep, descend=False) # Only allow one level of nesting
                        if thisdf is not None:
                            df[k] = thisdf
        return df

    def plot(self, style='fancy', fig_kw=None, plot_kw=None):
        """ Plot all the results """
        # Prepare the inputs
        fig_kw = sc.mergedicts(fig_kw)
        plot_kw = sc.mergedicts(plot_kw)
        results = list(self.all_results)
        nrows,ncols = sc.getrowscols(len(results))

        # Do the plotting
        with sc.options.with_style(style):
            fig = plt.figure(**fig_kw)
            for i,res in enumerate(results):
                ax = plt.subplot(nrows, ncols, i+1)
                res.plot(ax=ax, **plot_kw)
            sc.figlayout()
        return fig






"""
Import all Starsim modules
"""

# Start imports: version and settings
from .version import __version__, __versiondate__, __license__
from .settings import dtypes, options

# Optionally print the license
if options.license:
    print(__license__)

# Finish imports
from .utils         import *
from .arrays        import *
from .time          import *
from .parameters    import *
from .distributions import *
from .people        import *
from .modules       import *
from .networks      import *
from .results       import *
from .demographics  import *
from .products      import *
from .interventions import *
from .demographics  import *
from .disease       import *
from .diseases      import *
from .loop          import *
from .sim           import *
from .run           import *
from .calibration   import *
from .samples       import *

# Assign the root folder
import sciris as sc
root = sc.thispath(__file__).parent

# Double-check key requirements -- should match pyproject.toml
reqs = ['sciris>=3.2.0', 'pandas>=2.0.0']
msg = f'The following dependencies for Starsim {__version__} were not met: <MISSING>.'
sc.require(reqs, message=msg)
del sc, reqs, msg # Don't keep this in the module
